from __future__ import annotations

import asyncio
import logging
import queue
from datetime import UTC, datetime
from typing import Any

import pytest

from elke27_lib import client as client_mod
from elke27_lib import discovery, linking
from elke27_lib.client import Elke27Client, Result
from elke27_lib.errors import (
    CryptoError,
    E27AuthFailed,
    E27Error,
    E27LinkInvalid,
    E27NotReady,
    E27ProtocolError,
    E27ProvisioningRequired,
    E27Timeout,
    E27TransportError,
    Elke27AuthError,
    Elke27ConnectionError,
    Elke27CryptoError,
    Elke27InvalidArgument,
    Elke27LinkRequiredError,
    Elke27PermissionError,
    Elke27PinRequiredError,
    Elke27TimeoutError,
    InvalidPinError,
    NotAuthenticatedError,
    PanelNotDisarmedError,
    ProtocolError,
)
from elke27_lib.errors import (
    Elke27ProtocolError as Elke27ProtocolErrorV2,
)
from elke27_lib.events import (
    AreaStatusUpdated,
    AreaTroublesUpdated,
    ConnectionStateChanged,
    CsmSnapshotUpdated,
)
from elke27_lib.kernel import E27Kernel, KernelError, KernelMissingContextError
from elke27_lib.permissions import PermissionLevel
from elke27_lib.session import SessionConfig, SessionNotReadyError, SessionProtocolError
from elke27_lib.states import AreaState, OutputState, ZoneState
from elke27_lib.types import ArmMode, CsmSnapshot, LinkKeys


def _identity() -> linking.E27Identity:
    return linking.E27Identity("mn", "sn", "fw", "hw", "os")


def _event_base(kind: str) -> dict[str, Any]:
    return dict(
        kind=kind,
        at=0.0,
        seq=None,
        classification="LOCAL",
        route=("__local__", kind),
        session_id=1,
    )


def test_coerce_identity_and_link_keys() -> None:
    client = Elke27Client(kernel=E27Kernel())
    ident = client._coerce_identity(None)
    assert ident.mn and ident.sn
    identity = _identity()
    assert client._coerce_identity(identity) is identity
    with pytest.raises(Elke27InvalidArgument):
        client._coerce_identity(123)
    ident = client._coerce_identity({"mn": "m", "sn": "s"})
    assert ident.mn == "m"

    class _Val:
        def __bool__(self) -> bool:
            return True

        def __str__(self) -> str:
            return ""

    class _Map(dict):
        def get(self, key, default=None):  # type: ignore[no-untyped-def]
            if key in ("mn", "sn"):
                return _Val()
            return super().get(key, default)

    with pytest.raises(Elke27InvalidArgument):
        client._coerce_identity(_Map())

    keys = LinkKeys(tempkey_hex="aa", linkkey_hex="bb", linkhmac_hex="cc")
    coerced = client._coerce_link_keys(keys)
    assert coerced.linkkey_hex == "bb"


def test_raise_v2_error_mapping() -> None:
    client = Elke27Client(kernel=E27Kernel())
    with pytest.raises(Elke27LinkRequiredError):
        client._raise_v2_error(E27ProvisioningRequired("x"), phase="p")
    with pytest.raises(Elke27InvalidArgument):
        client._raise_v2_error(KernelMissingContextError("x"), phase="p")
    with pytest.raises(Elke27AuthError):
        client._raise_v2_error(E27AuthFailed("x"), phase="p")
    with pytest.raises(Elke27CryptoError):
        client._raise_v2_error(E27LinkInvalid("x"), phase="p")
    with pytest.raises(Elke27CryptoError):
        client._raise_v2_error(CryptoError("x"), phase="p")
    with pytest.raises(Elke27ProtocolErrorV2):
        client._raise_v2_error(E27ProtocolError("x"), phase="p")
    with pytest.raises(Elke27ConnectionError):
        client._raise_v2_error(E27TransportError("x"), phase="p")
    with pytest.raises(Elke27ConnectionError):
        client._raise_v2_error(TimeoutError(), phase="p")
    with pytest.raises(Elke27ConnectionError):
        client._raise_v2_error(TimeoutError("x"), phase="p")
    with pytest.raises(Elke27ProtocolErrorV2):
        client._raise_v2_error(RuntimeError("x"), phase="p")
    with pytest.raises(Elke27ProtocolErrorV2):
        client._raise_v2_error(KernelError("x"), phase="p")


def test_raise_v2_command_error_mapping() -> None:
    client = Elke27Client(kernel=E27Kernel())
    with pytest.raises(Elke27LinkRequiredError):
        client._raise_v2_command_error(E27ProvisioningRequired("x"))
    with pytest.raises(Elke27PermissionError):
        client._raise_v2_command_error(PanelNotDisarmedError("x"))
    with pytest.raises(Elke27PermissionError):
        client._raise_v2_command_error(NotAuthenticatedError("x"))
    with pytest.raises(Elke27AuthError):
        client._raise_v2_command_error(InvalidPinError("x"))
    with pytest.raises(Elke27TimeoutError):
        client._raise_v2_command_error(E27Timeout("x"))
    with pytest.raises(Elke27ConnectionError):
        client._raise_v2_command_error(E27NotReady("x"))
    with pytest.raises(Elke27CryptoError):
        client._raise_v2_command_error(CryptoError("x"))
    with pytest.raises(Elke27ProtocolErrorV2):
        client._raise_v2_command_error(E27ProtocolError("x"))
    with pytest.raises(Elke27ProtocolErrorV2):
        client._raise_v2_command_error(RuntimeError("x"))


def test_arm_mode_and_block_count_helpers() -> None:
    client = Elke27Client(kernel=E27Kernel())
    assert client._arm_mode_from_string("disarm") is ArmMode.DISARMED
    assert client._arm_mode_from_string("stay") is ArmMode.ARMED_STAY
    assert client._arm_mode_from_string("away") is ArmMode.ARMED_AWAY
    assert client._arm_mode_from_string("night") is ArmMode.ARMED_NIGHT
    assert client._arm_mode_from_string("foo") is None
    assert client._arm_mode_from_string(None) is None
    assert client._coerce_block_count("2") == 2
    assert client._coerce_block_count(0) is None


def test_build_maps_and_snapshot() -> None:
    kernel = E27Kernel()
    kernel.state.areas[1] = AreaState(area_id=1, name="A", arm_state="disarmed", ready=True)
    kernel.state.zones[1] = ZoneState(
        zone_id=1,
        name="Z",
        violated=True,
        bypassed=True,
        trouble=False,
        alarm=False,
        tamper=False,
        low_battery=True,
        attribs={"zone_type": "door", "kind": "entry"},
        definition=1,
    )
    kernel.state.zone_defs_by_id = {1: {"definition": "front"}}
    kernel.state.outputs[1] = OutputState(output_id=1, name="O", on=True)
    client = Elke27Client(kernel=kernel)
    assert client._build_area_map()[1].name == "A"
    assert client._build_zone_map()[1].bypassed is True
    assert client._build_zone_definitions()[1].definition == "front"
    assert client._build_output_map()[1].state is True
    client._replace_snapshot(
        panel_info=client._build_panel_info(),
        table_info=client._build_table_info(),
        areas=client._build_area_map(),
        zones=client._build_zone_map(),
        zone_definitions=client._build_zone_definitions(),
        outputs=client._build_output_map(),
        output_definitions=client._build_output_definitions(),
    )
    assert client.snapshot.version == 1


def test_queue_helpers() -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._enqueue_event(
        client_mod.Elke27Event(
            event_type=client_mod.EventType.SYSTEM,
            data={},
            seq=1,
            timestamp=datetime.now(UTC),
            raw_type="x",
        )
    )
    client._signal_event_stream_end()
    assert client._event_queue.qsize() >= 1


def test_handle_kernel_event_connection_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    called = {"safe": 0}
    monkeypatch.setattr(
        client, "_safe_request", lambda *_a, **_k: called.__setitem__("safe", called["safe"] + 1)
    )

    client._last_disconnect_at = 0.0
    client._now_monotonic = lambda: 10.0
    evt = ConnectionStateChanged(**_event_base(ConnectionStateChanged.KIND), connected=True)
    client._handle_kernel_event(evt)
    assert called["safe"] == 1

    evt = ConnectionStateChanged(**_event_base(ConnectionStateChanged.KIND), connected=False)
    client._handle_kernel_event(evt)
    assert client._last_disconnect_at is not None


def test_handle_kernel_event_area_paths(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    kernel = E27Kernel()
    kernel.state.areas[1] = AreaState(area_id=1, num_bypassed_zones=1)
    client = Elke27Client(kernel=kernel)
    called = {"refresh": 0}
    monkeypatch.setattr(
        client,
        "_refresh_all_zone_statuses_for_bypass_change",
        lambda _area: called.__setitem__("refresh", called["refresh"] + 1),
    )
    evt = AreaStatusUpdated(**_event_base(AreaStatusUpdated.KIND), area_id=1, changed_fields=())
    client._handle_kernel_event(evt)
    assert called["refresh"] == 1

    caplog.set_level(logging.WARNING, logger="elke27_lib.client")
    evt = AreaStatusUpdated(
        **_event_base(AreaStatusUpdated.KIND),
        area_id=1,
        changed_fields=("num_bypassed_zones",),
    )
    client._handle_kernel_event(evt)
    assert "bypass count changed" in caplog.text

    trouble_base = _event_base(AreaTroublesUpdated.KIND)
    trouble_base["classification"] = "BROADCAST"
    evt = AreaTroublesUpdated(
        **trouble_base,
        area_id=1,
        troubles=(),
    )
    client._handle_kernel_event(evt)
    assert called["refresh"] >= 2


def test_handle_kernel_event_csm_snapshot(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._awaiting_reconnect_csm_check = True
    now = datetime.now(UTC)
    client._reconnect_csm_snapshot = CsmSnapshot(
        domain_csms={"a": 1}, table_csms={}, version=1, updated_at=now
    )
    changed = CsmSnapshot(domain_csms={"a": 2}, table_csms={}, version=2, updated_at=now)
    called = {"safe": 0}
    monkeypatch.setattr(
        client, "_safe_request", lambda *_a, **_k: called.__setitem__("safe", called["safe"] + 1)
    )
    evt = CsmSnapshotUpdated(**_event_base(CsmSnapshotUpdated.KIND), snapshot=changed)
    client._handle_kernel_event(evt)
    assert called["safe"] == 1


def test_subscriber_error_logging(caplog: pytest.LogCaptureFixture) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    caplog.set_level(logging.WARNING, logger="elke27_lib.client")

    def _bad(_evt):  # type: ignore[no-untyped-def]
        raise ValueError("boom")

    client.subscribe(_bad)
    evt = ConnectionStateChanged(**_event_base(ConnectionStateChanged.KIND), connected=True)
    client._handle_kernel_event(evt)
    assert "Subscriber callback failed" in caplog.text


def test_on_kernel_event_dispatch() -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    called = {"count": 0}

    class _Loop:
        def call_soon_threadsafe(self, fn, *args):  # type: ignore[no-untyped-def]
            called["count"] += 1
            fn(*args)

    client._event_loop = _Loop()  # type: ignore[assignment]
    evt = ConnectionStateChanged(**_event_base(ConnectionStateChanged.KIND), connected=True)
    client._on_kernel_event(evt)
    assert called["count"] == 1


@pytest.mark.asyncio
async def test_async_discover_link_connect_disconnect(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)

    async def _discover(**_k):  # type: ignore[no-untyped-def]
        return client_mod.DiscoverResult(
            panels=[
                discovery.E27System(
                    panel_mac="m",
                    panel_host="h",
                    panel_name="n",
                    panel_serial=None,
                    port=2101,
                    tls_port=2102,
                )
            ]
        )

    monkeypatch.setattr(client_mod.E27Kernel, "discover", _discover)
    panels = await client.async_discover()
    assert panels and panels[0].host == "h"

    async def _link(*_a, **_k):  # type: ignore[no-untyped-def]
        return linking.E27LinkKeys("aa", "bb", "cc")

    monkeypatch.setattr(kernel, "link", _link)
    keys = await client.async_link(
        "h",
        2101,
        access_code="1",
        passphrase="2",
        client_identity={"mn": "m", "sn": "s"},
    )
    assert keys.linkkey_hex == "bb"

    async def _connect(*_a, **_k):  # type: ignore[no-untyped-def]
        return None

    monkeypatch.setattr(kernel, "connect", _connect)
    await client.async_connect("h", 2101, keys)
    assert client._connected is True

    async def _close():  # type: ignore[no-untyped-def]
        return None

    monkeypatch.setattr(kernel, "close", _close)
    await client.async_disconnect()
    assert client._connected is False


@pytest.mark.asyncio
async def test_async_refresh_csm(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._connected = True
    kernel.state.panel.connected = True
    now = datetime.now(UTC)
    snapshot = CsmSnapshot(domain_csms={"a": 1}, table_csms={}, version=1, updated_at=now)

    def _subscribe(cb, kinds=None):  # type: ignore[no-untyped-def]
        cb(CsmSnapshotUpdated(**_event_base(CsmSnapshotUpdated.KIND), snapshot=snapshot))
        return 1

    monkeypatch.setattr(kernel, "subscribe", _subscribe)
    monkeypatch.setattr(kernel, "unsubscribe", lambda _t: None)
    monkeypatch.setattr(kernel, "request_csm_refresh", lambda **_k: None)
    assert await client.async_refresh_csm() == snapshot


@pytest.mark.asyncio
async def test_async_set_output_and_arm(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._connected = True
    kernel.state.panel.connected = True

    async def _ok_execute(*_a, **_k):  # type: ignore[no-untyped-def]
        return Result(ok=True, data={}, error=None)

    monkeypatch.setattr(client, "async_execute", _ok_execute)
    await client.async_set_output(1, on=True)

    with pytest.raises(Elke27InvalidArgument):
        await client.async_arm_area(0, mode=ArmMode.ARMED_AWAY, pin="1")
    with pytest.raises(Elke27InvalidArgument):
        await client.async_disarm_area(1, pin="")


def test_request_authenticate_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    bad = client._request_authenticate(("control", "authenticate"), opaque=object())
    assert bad.ok is False

    q: queue.Queue[dict[str, object]] = queue.Queue(maxsize=1)
    q.put({"success": True})
    monkeypatch.setattr(kernel, "request", lambda *_a, **_k: 5)
    res = client._request_authenticate(("control", "authenticate"), opaque=q, pin="1234")
    assert res.ok is True and res.data == 5

    q = queue.Queue(maxsize=1)
    q.put({"success": False, "error_code": 5})
    res = client._request_authenticate(("control", "authenticate"), opaque=q, pin="1234")
    assert res.ok is False


def test_misc_helpers() -> None:
    client = Elke27Client(kernel=E27Kernel())
    assert client._has_expected_payload({"zone": {"get": {}}}, ("zone", "get")) is True
    assert client._extract_error_code({"zone": {"error_code": 5}}, ("zone", "get")) == 5
    payload = client._extract_response_payload({"zone": {"get": {"x": 1}}}, ("zone", "get"))
    assert payload["x"] == 1

    assert client._enforce_permissions("x", PermissionLevel.PLT_ENCRYPTION_KEY) is not None
    client._kernel.state.panel.session_id = 1
    assert client._enforce_permissions("x", PermissionLevel.PLT_ENCRYPTION_KEY) is None


def test_result_and_filtered_mapping() -> None:
    ok = client_mod.Result.success(1)
    assert ok.unwrap() == 1
    err = client_mod.Result.failure(ValueError("x"))
    with pytest.raises(ValueError):
        err.unwrap()
    with pytest.raises(E27Error):
        client_mod.Result(ok=True, data=None, error=None).unwrap()
    with pytest.raises(E27Error):
        client_mod.Result(ok=False, data=None, error=None).unwrap()

    mapping = client_mod._FilteredMapping({1: "a", 2: "b"}, {2})
    assert list(mapping) == [2]
    assert len(mapping) == 1
    assert mapping[2] == "b"
    with pytest.raises(KeyError):
        _ = mapping[1]

    state = E27Kernel().state
    state.table_info_by_domain["zone"] = {"table_elements": 2}
    assert list(client_mod._configured_ids_from_table(state, "zone")) == [1, 2]
    assert client_mod._table_elements_for_domain(state, "zone") == 2

    assert client_mod._configured_ids_from_table(state, "missing") == ()
    state.table_info_by_domain["bad"] = {"table_elements": 0}
    assert client_mod._configured_ids_from_table(state, "bad") == ()


@pytest.mark.asyncio
async def test_async_execute_single_and_paged(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._kernel.state.panel.session_id = 1

    class _Pending:
        def __init__(self) -> None:
            self.futures: dict[int, asyncio.Future[dict[str, Any]]] = {}

        def create(self, seq, **_k):  # type: ignore[no-untyped-def]
            fut = asyncio.get_running_loop().create_future()
            self.futures[seq] = fut
            return fut

        def drop(self, seq):  # type: ignore[no-untyped-def]
            self.futures.pop(seq, None)

    pending = _Pending()
    kernel._pending_responses = pending  # type: ignore[attr-defined]
    kernel.next_seq = lambda: 1  # type: ignore[assignment]
    kernel.register_sent_event = lambda _s, event: event.set()  # type: ignore[assignment]

    def _send(seq, domain, command, payload, **_k):  # type: ignore[no-untyped-def]
        msg = {domain: {command: {"ok": True}}}
        pending.futures[seq].set_result(msg)

    kernel.send_request_with_seq = _send  # type: ignore[assignment]

    def _gen(**_k: Any) -> tuple[dict[str, object], tuple[str, str]]:
        return {"x": 1}, ("zone", "get")

    spec = client_mod.CommandSpec(
        key="fake_get",
        domain="zone",
        command="get",
        generator=_gen,
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY,
    )
    monkeypatch.setitem(client_mod.COMMANDS, "fake_get", spec)
    monkeypatch.setattr(
        client_mod, "permission_for_generator", lambda *_a, **_k: PermissionLevel.PLT_ENCRYPTION_KEY
    )

    result = await client.async_execute("fake_get")
    assert result.ok is True

    def _gen_paged(block_id: int) -> tuple[dict[str, object], tuple[str, str]]:
        return {"block_id": block_id}, ("zone", "get_configured")

    paged = client_mod.CommandSpec(
        key="fake_paged",
        domain="zone",
        command="get_configured",
        generator=_gen_paged,
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY,
        response_mode="paged_blocks",
        block_field="block_id",
        block_count_field="block_count",
        merge_strategy=lambda blocks, total: {"blocks": len(blocks), "block_count": total},
    )
    monkeypatch.setitem(client_mod.COMMANDS, "fake_paged", paged)

    seq_counter = {"val": 0}

    def _next_seq():  # type: ignore[no-untyped-def]
        seq_counter["val"] += 1
        return seq_counter["val"]

    kernel.next_seq = _next_seq  # type: ignore[assignment]

    def _send_paged(seq, domain, command, payload, **_k):  # type: ignore[no-untyped-def]
        block_id = payload["block_id"]
        msg = {domain: {command: {"block_id": block_id, "block_count": 2}}}
        pending.futures[seq].set_result(msg)

    kernel.send_request_with_seq = _send_paged  # type: ignore[assignment]
    result = await client.async_execute("fake_paged")
    assert result.ok is True


def test_iter_causes_and_bootstrap() -> None:
    err = RuntimeError("root")
    err.__cause__ = ValueError("child")
    assert [type(e) for e in client_mod._iter_causes(err)] == [RuntimeError, ValueError]

    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    assert client.bootstrap_complete_counts is False


def test_init_config_and_roles() -> None:
    config = client_mod.ClientConfig(logger_name="elke27.test", event_queue_maxlen=5)
    client = Elke27Client(config=config)
    client.set_authenticated_role("master")
    assert client._auth_role == "master"


@pytest.mark.asyncio
async def test_wait_ready_and_subscribe() -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._ready_event.set()
    assert await client.wait_ready(0.01) is True

    called = {"count": 0}

    def _cb(_evt):  # type: ignore[no-untyped-def]
        called["count"] += 1

    unsub = client.subscribe(_cb)
    unsub2 = client.subscribe(_cb)
    assert unsub2() is True
    assert client.unsubscribe(_cb) is False
    assert unsub() is False

    unsub_typed = client.subscribe_typed(_cb)
    unsub_typed2 = client.subscribe_typed(_cb)
    assert unsub_typed2() is True
    assert client.unsubscribe_typed(_cb) is False
    assert unsub_typed() is False


def test_normalize_error_and_context(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    err = client._normalize_error(KernelMissingContextError("x"), phase="p")
    assert isinstance(err, E27Error)

    err = client._normalize_error(SessionNotReadyError("x"), phase="p")
    assert isinstance(err, E27NotReady)

    err = client._normalize_error(SessionProtocolError("x"), phase="p")
    assert isinstance(err, ProtocolError)

    class _Sess:
        cfg = SessionConfig(host="h", port=1)

    kernel._session = _Sess()  # type: ignore[assignment]
    ctx = client._error_context(phase="x", detail="y")
    assert ctx.host == "h" and ctx.port == 1


def test_payload_helpers() -> None:
    client = Elke27Client(kernel=E27Kernel())
    assert client._extract_error_code({"zone": 1}, ("zone", "get")) is None
    assert client._has_expected_payload({"zone": {"error_code": 1}}, ("zone", "get")) is True
    assert client._has_expected_payload({"zone": {}}, ("zone", "__root__")) is True
    payload = client._extract_response_payload({"zone": {"get": 5}}, ("zone", "get"))
    assert payload["value"] == 5


def test_pin_coercion_and_merge_helpers() -> None:
    client = Elke27Client(kernel=E27Kernel())

    def _gen(pin):  # type: ignore[no-untyped-def]
        return {"pin": pin}, ("x", "y")

    _gen.__annotations__ = {"pin": int}

    spec = client_mod.CommandSpec(
        key="x",
        domain="x",
        command="y",
        generator=_gen,
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY,
    )
    coerced = client._coerce_pin_for_generator(spec, {"pin": "123"})
    assert coerced["pin"] == 123

    assert (
        client._resolve_merge_strategy("output_configured") is client_mod._merge_configured_outputs
    )
    assert (
        client._resolve_merge_strategy("output_all_status")
        is client_mod._merge_output_status_strings
    )
    assert client._resolve_merge_strategy(None) is None

    blocks = [
        client_mod.PagedBlock(block_id=1, payload={"status": "A"}),
        client_mod.PagedBlock(block_id=2, payload={"status": "B"}),
    ]
    assert client_mod._merge_output_status_strings(blocks, 2)["status"] == "AB"

    blocks = [
        client_mod.PagedBlock(block_id=1, payload={"outputs": [1, 2]}),
        client_mod.PagedBlock(block_id=2, payload={"outputs": [2, 3]}),
    ]
    merged = client_mod._merge_configured_outputs(blocks, 2)
    assert merged["outputs"] == [1, 2, 3]


def test_all_areas_disarmed() -> None:
    kernel = E27Kernel()
    kernel.state.areas[1] = AreaState(area_id=1, arm_state="disarmed")
    client = Elke27Client(kernel=kernel)
    assert client._all_areas_disarmed() is True
    kernel.state.areas[2] = AreaState(area_id=2, arm_state="armed_away")
    assert client._all_areas_disarmed() is False


@pytest.mark.asyncio
async def test_async_connect_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())
    with pytest.raises(Elke27InvalidArgument):
        await client.async_connect("", 1, LinkKeys("a", "b", "c"))

    async def _connect(*_a, **_k):  # type: ignore[no-untyped-def]
        raise E27TransportError("x")

    monkeypatch.setattr(client._kernel, "connect", _connect)
    with pytest.raises(Elke27ConnectionError):
        await client.async_connect("h", 1, LinkKeys("a", "b", "c"))


@pytest.mark.asyncio
async def test_async_disconnect_error(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())

    async def _close():  # type: ignore[no-untyped-def]
        raise E27TransportError("x")

    monkeypatch.setattr(client._kernel, "close", _close)
    with pytest.raises(Elke27ConnectionError):
        await client.async_disconnect()


@pytest.mark.asyncio
async def test_refresh_domain_config_errors() -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._connected = True
    client._kernel.state.panel.connected = True
    with pytest.raises(Elke27InvalidArgument):
        await client.async_refresh_domain_config("")
    with pytest.raises(Elke27InvalidArgument):
        await client.async_refresh_domain_config("unknown")


def test_mark_inventory_and_status(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    kernel.state.inventory.configured_areas = {1}
    kernel.state.inventory.configured_zones = {2}
    kernel.state.inventory.configured_outputs = {3}
    called = {"queue": 0, "request": 0}
    monkeypatch.setattr(
        client,
        "_queue_bootstrap_attribs",
        lambda *_a: called.__setitem__("queue", called["queue"] + 1),
    )
    monkeypatch.setattr(
        client,
        "_request_initial_statuses",
        lambda *_a: called.__setitem__("request", called["request"] + 1),
    )
    client._mark_inventory_ready("area")
    assert called["queue"] == 1 and called["request"] == 1
    client._mark_inventory_ready("area")
    assert called["queue"] == 1


def test_refresh_helpers(monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture) -> None:
    kernel = E27Kernel()
    kernel.state.zones[1] = ZoneState(zone_id=1, area_id=1, bypassed=True)
    kernel.state.zones[2] = ZoneState(zone_id=2, area_id=1, bypassed=False)
    client = Elke27Client(kernel=kernel)
    called = {"count": 0}
    monkeypatch.setattr(
        client, "_safe_request", lambda *_a, **_k: called.__setitem__("count", called["count"] + 1)
    )
    caplog.set_level(logging.DEBUG, logger="elke27_lib.client")
    client._refresh_bypassed_zones_for_area(1)
    client._refresh_unbypassed_zones_for_area(1)
    assert called["count"] == 2
    client._refresh_all_zone_statuses_for_bypass_change(1)
    assert called["count"] == 3


def test_record_and_suppress_bypass() -> None:
    kernel = E27Kernel()
    kernel.state.zones[1] = ZoneState(zone_id=1, area_id=1, bypassed=True)
    client = Elke27Client(kernel=kernel)
    client._record_local_zone_bypass(1)
    assert client._should_suppress_area_bypass_refresh(1) is True
    client._pending_bypass_by_area[1] = 0.0
    kernel.now = lambda: 10.0  # type: ignore[assignment]
    assert client._should_suppress_area_bypass_refresh(1) is False


def test_mark_status_seen() -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._status_pending["zone"] = {1, 2}
    client._mark_status_seen("zone", [1, 2])
    assert client._status_ready["zone"] is True


def test_on_kernel_event_no_loop() -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._event_loop = None
    evt = ConnectionStateChanged(**_event_base(ConnectionStateChanged.KIND), connected=True)
    client._on_kernel_event(evt)


@pytest.mark.asyncio
async def test_async_discover_and_link_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())

    async def _discover(*_a, **_k):  # type: ignore[no-untyped-def]
        raise E27AuthFailed("x")

    monkeypatch.setattr(client_mod.E27Kernel, "discover", _discover)
    monkeypatch.setattr(client, "_raise_v2_error", lambda *_a, **_k: None)
    with pytest.raises(AssertionError):
        await client.async_discover()

    with pytest.raises(Elke27InvalidArgument):
        await client.async_link(
            "", 1, access_code="1", passphrase="2", client_identity={"mn": "m", "sn": "s"}
        )
    with pytest.raises(Elke27InvalidArgument):
        await client.async_link(
            "h", 0, access_code="1", passphrase="2", client_identity={"mn": "m", "sn": "s"}
        )
    with pytest.raises(Elke27InvalidArgument):
        await client.async_link(
            "h", 1, access_code="", passphrase="2", client_identity={"mn": "m", "sn": "s"}
        )
    with pytest.raises(Elke27InvalidArgument):
        await client.async_link(
            "h", 1, access_code="1", passphrase="", client_identity={"mn": "m", "sn": "s"}
        )
    with pytest.raises(Elke27InvalidArgument):
        await client.async_link("h", 1, access_code="1", passphrase="2", client_identity=None)

    async def _link(*_a, **_k):  # type: ignore[no-untyped-def]
        raise E27AuthFailed("x")

    monkeypatch.setattr(client._kernel, "link", _link)
    with pytest.raises(AssertionError):
        await client.async_link(
            "h", 1, access_code="1", passphrase="2", client_identity={"mn": "m", "sn": "s"}
        )


def test_request_authenticate_timeout_and_invalid(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())
    q: queue.Queue[dict[str, object]] = queue.Queue(maxsize=1)
    monkeypatch.setattr(client._kernel, "request", lambda *_a, **_k: 5)
    result = client._request_authenticate(("control", "authenticate"), opaque=q, pin="1234")
    assert result.ok is False and isinstance(result.error, E27Timeout)

    q = queue.Queue(maxsize=1)
    q.put("bad")  # type: ignore[arg-type]
    result = client._request_authenticate(("control", "authenticate"), opaque=q, pin="1234")
    assert result.ok is False and isinstance(result.error, ProtocolError)


@pytest.mark.asyncio
async def test_async_authenticate_timeout(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._kernel.state.panel.session_id = 1

    class _Pending:
        def create(self, *_a, **_k):  # type: ignore[no-untyped-def]
            fut = asyncio.get_running_loop().create_future()
            return fut

        def drop(self, *_a, **_k):  # type: ignore[no-untyped-def]
            return None

    client._kernel._pending_responses = _Pending()  # type: ignore[attr-defined]
    client._kernel.register_sent_event = lambda _s, event: event.set()  # type: ignore[assignment]
    client._kernel.send_request_with_seq = lambda *_a, **_k: None  # type: ignore[assignment]
    res = await client._async_authenticate(pin=1234, timeout_s=0.0)
    assert res.ok is False and isinstance(res.error, E27Timeout)


@pytest.mark.asyncio
async def test_async_execute_permission_and_disarmed(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._kernel.state.panel.session_id = 1
    client._kernel.state.areas[1] = AreaState(area_id=1, arm_state="armed_away")
    monkeypatch.setattr(
        client_mod,
        "permission_for_generator",
        lambda *_a, **_k: PermissionLevel.PLT_ENCRYPTION_KEY_DISARMED,
    )

    spec = client_mod.CommandSpec(
        key="needs_disarmed",
        domain="zone",
        command="get",
        generator=lambda **_k: ({}, ("zone", "get")),  # type: ignore[no-untyped-def]
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY_DISARMED,
    )
    monkeypatch.setitem(client_mod.COMMANDS, "needs_disarmed", spec)
    result = await client.async_execute("needs_disarmed")
    assert isinstance(result.error, Elke27PermissionError)


@pytest.mark.asyncio
async def test_async_execute_pin_validation(monkeypatch: pytest.MonkeyPatch) -> None:
    client = Elke27Client(kernel=E27Kernel())
    client._kernel.state.panel.session_id = 1
    monkeypatch.setattr(
        client_mod, "permission_for_generator", lambda *_a, **_k: PermissionLevel.PLT_ANY_USER
    )

    spec = client_mod.CommandSpec(
        key="needs_pin",
        domain="area",
        command="set",
        generator=lambda **_k: ({}, ("area", "set")),  # type: ignore[no-untyped-def]
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ANY_USER,
    )
    monkeypatch.setitem(client_mod.COMMANDS, "needs_pin", spec)
    result = await client.async_execute("needs_pin")
    assert isinstance(result.error, Elke27PinRequiredError)
    result = await client.async_execute("needs_pin", pin="abc")
    assert isinstance(result.error, InvalidPinError)


@pytest.mark.asyncio
async def test_async_execute_single_missing_payload(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._kernel.state.panel.session_id = 1
    monkeypatch.setattr(
        client_mod, "permission_for_generator", lambda *_a, **_k: PermissionLevel.PLT_ENCRYPTION_KEY
    )

    class _Pending:
        def __init__(self) -> None:
            self.futures: dict[int, asyncio.Future[dict[str, Any]]] = {}

        def create(self, seq, **_k):  # type: ignore[no-untyped-def]
            fut = asyncio.get_running_loop().create_future()
            self.futures[seq] = fut
            return fut

        def drop(self, seq):  # type: ignore[no-untyped-def]
            self.futures.pop(seq, None)

    pending = _Pending()
    kernel._pending_responses = pending  # type: ignore[attr-defined]
    kernel.next_seq = lambda: 1  # type: ignore[assignment]
    kernel.register_sent_event = lambda _s, event: event.set()  # type: ignore[assignment]

    def _send(seq, domain, command, payload, **_k):  # type: ignore[no-untyped-def]
        pending.futures[seq].set_result({domain: {}})

    kernel.send_request_with_seq = _send  # type: ignore[assignment]

    spec = client_mod.CommandSpec(
        key="fake_get_missing",
        domain="zone",
        command="get",
        generator=lambda **_k: ({}, ("zone", "get")),  # type: ignore[no-untyped-def]
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY,
    )
    monkeypatch.setitem(client_mod.COMMANDS, "fake_get_missing", spec)
    result = await client.async_execute("fake_get_missing")
    assert result.ok is False and isinstance(result.error, ProtocolError)


@pytest.mark.asyncio
async def test_async_execute_paged_block_mismatch(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    client._kernel.state.panel.session_id = 1
    monkeypatch.setattr(
        client_mod, "permission_for_generator", lambda *_a, **_k: PermissionLevel.PLT_ENCRYPTION_KEY
    )

    class _Pending:
        def __init__(self) -> None:
            self.futures: dict[int, asyncio.Future[dict[str, Any]]] = {}

        def create(self, seq, **_k):  # type: ignore[no-untyped-def]
            fut = asyncio.get_running_loop().create_future()
            self.futures[seq] = fut
            return fut

        def drop(self, seq):  # type: ignore[no-untyped-def]
            self.futures.pop(seq, None)

    pending = _Pending()
    kernel._pending_responses = pending  # type: ignore[attr-defined]
    kernel.register_sent_event = lambda _s, event: event.set()  # type: ignore[assignment]
    seq_counter = {"val": 0}
    kernel.next_seq = lambda: (
        seq_counter.__setitem__("val", seq_counter["val"] + 1) or seq_counter["val"]
    )  # type: ignore[assignment]

    def _send(seq, domain, command, payload, **_k):  # type: ignore[no-untyped-def]
        block_id = payload["block_id"]
        block_count = 2 if block_id == 1 else 3
        msg = {domain: {command: {"block_id": block_id, "block_count": block_count}}}
        pending.futures[seq].set_result(msg)

    kernel.send_request_with_seq = _send  # type: ignore[assignment]

    paged = client_mod.CommandSpec(
        key="fake_paged_mismatch",
        domain="zone",
        command="get_configured",
        generator=lambda block_id: ({"block_id": block_id}, ("zone", "get_configured")),  # type: ignore[no-untyped-def]
        handler=lambda *_a, **_k: True,  # type: ignore[no-untyped-def]
        min_permission=PermissionLevel.PLT_ENCRYPTION_KEY,
        response_mode="paged_blocks",
        block_field="block_id",
        block_count_field="block_count",
        merge_strategy=lambda blocks, total: {"blocks": len(blocks), "block_count": total},
    )
    monkeypatch.setitem(client_mod.COMMANDS, "fake_paged_mismatch", paged)
    result = await client.async_execute("fake_paged_mismatch")
    assert result.ok is False and isinstance(result.error, ProtocolError)


def test_request_and_pump_once_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    kernel = E27Kernel()
    client = Elke27Client(kernel=kernel)
    monkeypatch.setattr(kernel, "request", lambda *_a, **_k: (_ for _ in ()).throw(E27Error("x")))
    result = client.request(("zone", "get_status"), zone_id=1)
    assert result.ok is False

    class _Sess:
        cfg = SessionConfig(host="h", port=1)

        def pump_once(self, **_k):  # type: ignore[no-untyped-def]
            raise SessionProtocolError("x")

    kernel._session = _Sess()  # type: ignore[assignment]
    result = client.pump_once()
    assert result.ok is False and isinstance(result.error, ProtocolError)



from __future__ import annotations

import pytest

from elke27_lib import presentation
from elke27_lib.errors import E27ProtocolError


def test_require_helpers_and_iv_length() -> None:
    with pytest.raises(E27ProtocolError):
        presentation._require_len("ciphertext", b"", 16)
    with pytest.raises(E27ProtocolError):
        presentation._require_key_16(b"\x00" * 15, context_phase="x")
    with pytest.raises(E27ProtocolError):
        presentation._aes128_cbc_decrypt(key=b"\x00" * 16, iv=b"\x00", ciphertext=b"\x00" * 16)
    with pytest.raises(E27ProtocolError):
        presentation._aes128_cbc_encrypt(key=b"\x00" * 16, iv=b"\x00", plaintext=b"\x00" * 16)


def test_decrypt_schema0_envelope_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_schema0_envelope(
            protocol_byte=0x01, ciphertext=b"\x00" * 16, session_key=b"\x00" * 16
        )

    monkeypatch.setattr(presentation, "swap_endianness", lambda b: b)
    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: b"\x00" * 8)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_schema0_envelope(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, session_key=b"\x00" * 16
        )

    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: b"\x00" * 16)
    monkeypatch.setattr(presentation, "protocol_padding_len", lambda _p: 16)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_schema0_envelope(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, session_key=b"\x00" * 16
        )

    monkeypatch.setattr(presentation, "protocol_padding_len", lambda _p: 3)
    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: b"\x00" * 10)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_schema0_envelope(
            protocol_byte=0x80 | 0x03, ciphertext=b"\x00" * 16, session_key=b"\x00" * 16
        )


def test_encrypt_schema0_envelope_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    with pytest.raises(E27ProtocolError):
        presentation.encrypt_schema0_envelope(payload=b"", session_key=b"\x00" * 16, src=256)
    with pytest.raises(E27ProtocolError):
        presentation.encrypt_schema0_envelope(
            payload=b"", session_key=b"\x00" * 16, envelope_seq=-1
        )

    monkeypatch.setattr(presentation, "calculate_block_padding", lambda _n: 16)
    with pytest.raises(E27ProtocolError):
        presentation.encrypt_schema0_envelope(payload=b"", session_key=b"\x00" * 16)

    monkeypatch.setattr(presentation, "calculate_block_padding", lambda _n: 1)
    with pytest.raises(E27ProtocolError):
        presentation.encrypt_schema0_envelope(payload=b"", session_key=b"\x00" * 16)


def test_decrypt_api_link_response_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_api_link_response(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, tempkey_hex="zz"
        )

    monkeypatch.setattr(presentation, "swap_endianness", lambda b: b)
    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: b"\x00" * 5)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_api_link_response(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, tempkey_hex="00" * 16
        )

    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: b"\x00" * 10)
    monkeypatch.setattr(presentation, "protocol_padding_len", lambda _p: 5)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_api_link_response(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, tempkey_hex="00" * 16
        )

    pt = b"\x00" * 8 + b"\x00\x00"
    monkeypatch.setattr(presentation, "protocol_padding_len", lambda _p: 0)
    monkeypatch.setattr(presentation, "_aes128_cbc_decrypt", lambda **_k: pt)
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_api_link_response(
            protocol_byte=0x80, ciphertext=b"\x00" * 16, tempkey_hex="00" * 16
        )


def test_decrypt_key_field_with_linkkey_hex_errors() -> None:
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_key_field_with_linkkey(linkkey_hex="zz", ciphertext_hex="00")
    with pytest.raises(E27ProtocolError):
        presentation.decrypt_key_field_with_linkkey(linkkey_hex="00" * 16, ciphertext_hex="zz")



"""
elke27_lib/handlers/user.py

Read-only handlers for the "user" domain.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    ApiError,
    AuthorizationRequiredEvent,
    Event,
    UserConfiguredInventoryReady,
)
from elke27_lib.states import PanelState, UserState

EmitFn = Callable[[Event, DispatchContext], None]
NowFn = Callable[[], float]


def _as_mapping(obj: object) -> Mapping[str, Any] | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, Any], obj)
    return None


def _coerce_int(value: object) -> int | None:
    return value if isinstance(value, int) else None


def make_user_get_configured_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("user","get_configured").
    """

    def handler_user_get_configured(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        user_obj = _as_mapping(msg.get("user"))
        if user_obj is None:
            return False

        payload = _as_mapping(user_obj.get("get_configured"))
        if payload is None:
            return False

        error_code = payload.get("error_code", user_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="user",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="user",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        users = _extract_configured_ids(payload)
        if users:
            state.inventory.configured_users.update(users)

        block_id = payload.get("block_id")
        block_count = payload.get("block_count")
        if (
            isinstance(block_id, int)
            and isinstance(block_count, int)
            and block_count >= 1
            and block_id >= block_count
        ):
            state.inventory.configured_users_complete = True
            emit(
                UserConfiguredInventoryReady(
                    kind=UserConfiguredInventoryReady.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                ),
                ctx,
            )

        state.panel.last_message_at = now()
        return True

    return handler_user_get_configured


def make_user_get_attribs_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("user","get_attribs").
    """

    def handler_user_get_attribs(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        user_obj = _as_mapping(msg.get("user"))
        if user_obj is None:
            return False

        payload = _as_mapping(user_obj.get("get_attribs"))
        if payload is None:
            return False

        error_code = payload.get("error_code", user_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="user",
                        entity_id=_coerce_int(payload.get("user_id")),
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="user",
                    entity_id=_coerce_int(payload.get("user_id")),
                    message=None,
                ),
                ctx,
            )
            return True

        user_id = payload.get("user_id")
        if not isinstance(user_id, int) or user_id < 1:
            return False

        user = state.get_or_create_user(user_id)
        changed: set[str] = set()
        _apply_user_attribs(user, payload, changed)
        user.last_update_at = now()
        state.panel.last_message_at = user.last_update_at

        return True

    return handler_user_get_attribs


def _extract_configured_ids(payload: Mapping[str, Any]) -> set[int]:
    for key in ("users", "user_ids", "configured_users", "configured_user_ids"):
        value = payload.get(key)
        if isinstance(value, list):
            ids: set[int] = set()
            items = cast(list[object], value)
            for item in items:
                if isinstance(item, int) and item >= 1:
                    ids.add(item)
            return ids
    return set()


def _normalize_name(value: Any) -> str | None:
    if value is None:
        return None
    text = str(value).strip()
    return text if text else None


def _apply_user_attribs(user: UserState, payload: Mapping[str, Any], changed: set[str]) -> None:
    if "name" in payload:
        name = _normalize_name(payload.get("name"))
        if user.name != name:
            user.name = name
            changed.add("name")
    if "group_id" in payload:
        group_id = payload.get("group_id")
        if isinstance(group_id, int) and user.group_id != group_id:
            user.group_id = group_id
            changed.add("group_id")
    if "enabled" in payload:
        enabled = payload.get("enabled")
        if isinstance(enabled, bool) and user.enabled != enabled:
            user.enabled = enabled
            changed.add("enabled")
    if "pin" in payload:
        pin = payload.get("pin")
        if isinstance(pin, int) and user.pin != pin:
            user.pin = pin
            changed.add("pin")
    if "flags" in payload:
        flags = payload.get("flags")
        if isinstance(flags, list) and user.flags != flags:
            user.flags = flags
            changed.add("flags")

    for key, value in payload.items():
        if key in {"user_id", "error_code", "name", "group_id", "enabled", "pin", "flags"}:
            continue
        if user.fields.get(key) != value:
            user.fields[key] = value
            changed.add(key)



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TypeVar

import pytest
from _pytest.monkeypatch import MonkeyPatch

from elke27_lib import ArmMode, ClientConfig, Elke27Client
from elke27_lib.client import Result
from elke27_lib.errors import (
    E27ProvisioningRequired,
    Elke27DisconnectedError,
    Elke27InvalidArgument,
    Elke27LinkRequiredError,
    Elke27PermissionError,
    PermissionDeniedError,
)


class _FakePanel:
    connected: bool

    def __init__(self, connected: bool = True) -> None:
        self.connected = connected


class _FakeState:
    panel: _FakePanel

    def __init__(self) -> None:
        self.panel = _FakePanel()


class _FakeKernel:
    sent: list[tuple[str, dict[str, object]]]
    state: _FakeState

    def __init__(self) -> None:
        self.sent = []
        self.state = _FakeState()

    def subscribe(
        self, _callback: Callable[[object], None], _kinds: Iterable[str] | None = None
    ) -> int:
        return 1

    def unsubscribe(self, _token: int) -> bool:
        return True


T = TypeVar("T")


def _ok(value: T) -> Result[T]:
    return Result(ok=True, data=value, error=None)


def _err(error: BaseException) -> Result[T]:
    return Result(ok=False, data=None, error=error)


@pytest.mark.asyncio
async def test_async_set_output_calls_kernel_and_no_snapshot_mutation():
    kernel = _FakeKernel()
    client = Elke27Client(kernel=kernel, config=ClientConfig(event_queue_size=1))
    client._connected = True
    before = client.snapshot

    async def _ok_execute(command_key: str, **params: object) -> Result[dict[str, object]]:
        kernel.sent.append((command_key, dict(params)))
        return _ok({})

    client.async_execute = _ok_execute

    await client.async_set_output(3, on=True)

    assert kernel.sent == [("output_set_status", {"output_id": 3, "status": "ON"})]
    assert client.snapshot is before


@pytest.mark.asyncio
async def test_async_set_output_requires_connection():
    kernel = _FakeKernel()
    kernel.state.panel.connected = False
    client = Elke27Client(kernel=kernel)
    with pytest.raises(Elke27DisconnectedError):
        await client.async_set_output(1, on=False)


@pytest.mark.asyncio
async def test_async_arm_disarm_calls_async_execute(monkeypatch: MonkeyPatch):
    client = Elke27Client()
    before = client.snapshot

    async def _ok_execute(*_args: object, **_kwargs: object) -> Result[dict[str, object]]:
        return _ok({})

    monkeypatch.setattr(client, "async_execute", _ok_execute)

    await client.async_arm_area(1, mode=ArmMode.ARMED_AWAY, pin="1234")
    await client.async_disarm_area(1, pin="1234")
    assert client.snapshot is before


@pytest.mark.asyncio
async def test_argument_validation():
    client = Elke27Client()
    with pytest.raises(Elke27InvalidArgument):
        await client.async_set_output(0, on=True)
    with pytest.raises(Elke27InvalidArgument):
        await client.async_arm_area(0, mode=ArmMode.ARMED_AWAY, pin="1234")
    with pytest.raises(Elke27InvalidArgument):
        await client.async_arm_area(1, mode=ArmMode.ARMED_NIGHT, pin="1234")
    with pytest.raises(Elke27InvalidArgument):
        await client.async_disarm_area(1, pin="")


@pytest.mark.asyncio
async def test_error_mapping_for_arm_disarm(monkeypatch: MonkeyPatch):
    client = Elke27Client()

    async def _link_required(*_args: object, **_kwargs: object) -> Result[dict[str, object]]:
        return _err(E27ProvisioningRequired("missing link"))

    monkeypatch.setattr(client, "async_execute", _link_required)
    with pytest.raises(Elke27LinkRequiredError):
        await client.async_arm_area(1, mode=ArmMode.ARMED_STAY, pin="1234")

    async def _perm_error(*_args: object, **_kwargs: object) -> Result[dict[str, object]]:
        return _err(PermissionDeniedError("no"))

    monkeypatch.setattr(client, "async_execute", _perm_error)
    with pytest.raises(Elke27PermissionError) as exc_info:
        await client.async_disarm_area(1, pin="1234")
    assert exc_info.value.is_transient is False



#!/usr/bin/env python3
"""
Minimal Elke27Client contract example.

Flow:
- instantiate client
- connect
- wait_ready
- subscribe for events
- receive at least one event
- read a snapshot

Environment variables:
  ELKE27_MN, ELKE27_SN, ELKE27_FWVER, ELKE27_HWVER, ELKE27_OSVER
  ELKE27_ACCESS_CODE, ELKE27_PASSPHRASE
"""

from __future__ import annotations

import asyncio
import logging
import os
import sys
import time
from dataclasses import dataclass

from elke27_lib.client import E27Identity, Elke27Client
from elke27_lib.events import (
    AreaConfiguredInventoryReady,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    ConnectionStateChanged,
    ZoneConfiguredInventoryReady,
)
from elke27_lib.session import SessionConfig

LOG = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class Credentials:
    access_code: str
    passphrase: str


def _env(name: str, default: str | None = None) -> str | None:
    return os.environ.get(name, default)

def _print_summary() -> None:
    summary_path = os.path.join(os.path.dirname(__file__), "..", "docs", "client_contract_summary.md")
    print("Client Contract Summary", flush=True)
    print("-----------------------", flush=True)
    try:
        with open(summary_path, encoding="utf-8") as fh:
            print(fh.read().strip(), flush=True)
    except OSError as exc:
        print(f"[summary unavailable: {exc}]", flush=True)
    print("", flush=True)


async def main_async() -> int:
    logging.basicConfig(level=logging.INFO, format="%(message)s", stream=sys.stdout)
    _print_summary()

    mn = _env("ELKE27_MN", "")
    sn = _env("ELKE27_SN", "")
    access_code = _env("ELKE27_ACCESS_CODE", "")
    passphrase = _env("ELKE27_PASSPHRASE", "")

    if not mn or not sn:
        raise SystemExit("ERROR: set ELKE27_MN and ELKE27_SN (client_identity).")
    if not access_code or not passphrase:
        raise SystemExit("ERROR: set ELKE27_ACCESS_CODE and ELKE27_PASSPHRASE (provisioning).")

    client = Elke27Client(filter_attribs_to_configured=True)
    print("filter_attribs_to_configured=True", flush=True)

    event_seen = asyncio.Event()
    auth_seen = asyncio.Event()
    area_auth_required = False
    zone_auth_required = False
    event_count = 0

    def on_event(evt) -> None:
        nonlocal event_count, area_auth_required, zone_auth_required
        event_count += 1
        if not hasattr(evt, "kind"):
            raise RuntimeError("Non-semantic event received.")
        if isinstance(
            evt,
            (
                ConnectionStateChanged,
                AuthorizationRequiredEvent,
                BootstrapCountsReady,
                AreaConfiguredInventoryReady,
                ZoneConfiguredInventoryReady,
            ),
        ):
            LOG.info("event=%s domain=%s", evt.kind, evt.domain)
        if isinstance(evt, BootstrapCountsReady):
            LOG.info("bootstrap_counts_ready: bootstrap_complete_counts=%s", client.bootstrap_complete_counts)
        if isinstance(evt, AreaConfiguredInventoryReady):
            LOG.info("area_configured_inventory_ready: configured_areas=%s", len(client.areas))
        if isinstance(evt, ZoneConfiguredInventoryReady):
            LOG.info("zone_configured_inventory_ready: configured_zones=%s", len(client.zones))
        event_seen.set()
        if isinstance(evt, AuthorizationRequiredEvent):
            auth_seen.set()
            if evt.scope == "area":
                area_auth_required = True
            if evt.scope == "zone":
                zone_auth_required = True

    client.subscribe(
        on_event,
        kinds=[
            ConnectionStateChanged.KIND,
            AuthorizationRequiredEvent.KIND,
            BootstrapCountsReady.KIND,
            AreaConfiguredInventoryReady.KIND,
            ZoneConfiguredInventoryReady.KIND,
        ],
    )

    discover = await client.discover()
    panels = discover.unwrap().panels
    if not panels:
        raise SystemExit("ERROR: no panels discovered.")

    client_identity = E27Identity(
        mn=mn,
        sn=sn,
        fwver=_env("ELKE27_FWVER", "0") or "0",
        hwver=_env("ELKE27_HWVER", "0") or "0",
        osver=_env("ELKE27_OSVER", "0") or "0",
    )
    creds = Credentials(access_code=access_code, passphrase=passphrase)

    link_keys = (await client.link(panels[0], client_identity, creds)).unwrap()
    session_cfg = SessionConfig(host=panels[0].panel_host, port=int(panels[0].port))
    LOG.info("import_path=elke27_lib.client.Elke27Client")
    LOG.info("is_ready(before)=%s", client.is_ready)
    LOG.info("bootstrap_complete_counts(before)=%s", client.bootstrap_complete_counts)
    (await client.connect(link_keys, client_identity=client_identity, session_config=session_cfg)).unwrap()

    await asyncio.wait_for(event_seen.wait(), timeout=5.0)

    client.request(("control", "get_version_info")).unwrap()
    client.request(("area", "get_table_info")).unwrap()
    client.request(("zone", "get_table_info")).unwrap()
    client.request(("output", "get_table_info")).unwrap()
    client.request(("tstat", "get_table_info")).unwrap()
    client.request(("network", "get_ssid")).unwrap()

    deadline = time.monotonic() + 30.0
    while time.monotonic() < deadline:
        await asyncio.to_thread(client.pump_once, timeout_s=0.25)
        if client.is_ready and auth_seen.is_set():
            break

    if not client.is_ready:
        LOG.info("table_info keys=%s", sorted(client.table_info.keys()))
        raise SystemExit("ERROR: client did not become ready.")

    panel_info = client.panel_info
    LOG.info("is_ready(after)=%s", client.is_ready)
    LOG.info("bootstrap_complete_counts(after)=%s", client.bootstrap_complete_counts)
    LOG.info("panel session_id=%s", panel_info.session_id)
    LOG.info("panel mac=%s", getattr(panel_info, "mac", None) or "MISSING")
    LOG.info("panel model=%s", panel_info.model or "MISSING")
    LOG.info("panel firmware=%s", panel_info.firmware or "MISSING")
    LOG.info("panel serial=%s", panel_info.serial or "MISSING")

    table_info = client.table_info
    zone_info = table_info.get("zone") if isinstance(table_info, dict) else None
    area_info = table_info.get("area") if isinstance(table_info, dict) else None
    output_info = table_info.get("output") if isinstance(table_info, dict) else None
    tstat_info = table_info.get("tstat") if isinstance(table_info, dict) else None
    zone_count = zone_info.get("table_elements") if isinstance(zone_info, dict) else None
    area_count = area_info.get("table_elements") if isinstance(area_info, dict) else None
    output_count = output_info.get("table_elements") if isinstance(output_info, dict) else None
    tstat_count = tstat_info.get("table_elements") if isinstance(tstat_info, dict) else None
    LOG.info("table_info.zone.table_elements=%s", zone_count if zone_count is not None else "MISSING")
    LOG.info("table_info.area.table_elements=%s", area_count if area_count is not None else "MISSING")
    LOG.info("table_info.output.table_elements=%s", output_count if output_count is not None else "MISSING")
    LOG.info("table_info.tstat.table_elements=%s", tstat_count if tstat_count is not None else "MISSING")

    configured_area_ids = list(client.areas)
    configured_zone_ids = list(client.zones)
    if area_auth_required and not configured_area_ids:
        LOG.info("configured_areas=MISSING (auth required)")
    else:
        LOG.info("configured_areas=%s", len(configured_area_ids))
    if zone_auth_required and not configured_zone_ids:
        LOG.info("configured_zones=MISSING (auth required)")
    else:
        LOG.info("configured_zones=%s", len(configured_zone_ids))

    for area_id in sorted(configured_area_ids)[:5]:
        area = client.areas.get(area_id)
        name = getattr(area, "name", None) if area is not None else None
        LOG.info("area[%s].name=%s", area_id, name or "MISSING")

    for zone_id in sorted(configured_zone_ids)[:5]:
        zone = client.zones.get(zone_id)
        name = getattr(zone, "name", None) if zone is not None else None
        LOG.info("zone[%s].name=%s", zone_id, name or "MISSING")

    if auth_seen.is_set():
        LOG.info("auth_required_event=seen")
    else:
        LOG.info("auth_required_event=not_seen")

    client.unsubscribe(on_event)
    prior_count = event_count
    await client.disconnect()
    await asyncio.sleep(1.0)
    if event_count != prior_count:
        LOG.info("unsubscribe_check=failed (events_after_unsubscribe=%s)", event_count - prior_count)
    else:
        LOG.info("unsubscribe_check=ok")

    return 0


def main() -> int:
    return asyncio.run(main_async())


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from elke27_lib.generators.registry import COMMANDS


def test_registry_handlers_use_canonical_names() -> None:
    for key, spec in COMMANDS.items():
        assert spec.handler.__name__ == f"handler_{key}"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

import pytest

from elke27_lib import redact


class _Color(Enum):
    RED = "red"


@dataclass
class _Inner:
    token: str
    value: int


def test_should_redact() -> None:
    assert redact._should_redact("passphrase") is True
    assert redact._should_redact("apiToken") is True
    assert redact._should_redact("secret_key") is True
    assert redact._should_redact("monkey") is False


def test_normalize_for_diagnostics() -> None:
    obj = _Inner(token="abc", value=3)
    out = redact._normalize_for_diagnostics(obj)
    assert out["token"] == "abc"
    assert out["value"] == 3
    assert redact._normalize_for_diagnostics(None) is None
    assert redact._normalize_for_diagnostics({"a": 1}) == {"a": 1}
    assert redact._normalize_for_diagnostics([b"\x00", _Color.RED]) == ["<bytes:1>", "red"]
    assert "object at" in redact._normalize_for_diagnostics(object())


def test_redact_for_diagnostics(monkeypatch: pytest.MonkeyPatch) -> None:
    data = {"passphrase": "secret", "safe": 1}
    out = redact.redact_for_diagnostics(data)
    assert out["passphrase"] == "***"
    assert out["safe"] == 1
    out_dc = redact.redact_for_diagnostics(_Inner(token="secret", value=2))
    assert out_dc["token"] == "***"
    assert out_dc["value"] == 2
    assert redact.redact_for_diagnostics(_Color.RED) == "red"
    assert redact.redact_for_diagnostics(b"\x00") == "<bytes:1>"

    monkeypatch.setattr(redact, "REDACT_DIAGNOSTICS", False)
    out2 = redact.redact_for_diagnostics(data)
    assert out2["passphrase"] == "secret"



#!/usr/bin/env python3
"""
e27_network_param_live.py

Live example: discover -> link -> connect -> network.get_ssid + network.get_rssi
with call-site authorization retry (PIN prompt).
"""

from __future__ import annotations

import argparse
import asyncio
import logging
import time
from dataclasses import dataclass

from elke27_lib import Elk, linking
from elke27_lib.events import ApiError, NetworkRssiUpdated, NetworkSsidResultsUpdated
from elke27_lib.session import SessionConfig

LOG = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class E27Credentials:
    access_code: str
    passphrase: str


def _env(name: str, default: str | None = None) -> str | None:
    import os
    value = os.environ.get(name, default)
    if value == "":
        return default
    return value


def _env_any(names: list[str], default: str | None = None) -> str | None:
    for name in names:
        value = _env(name)
        if value:
            return value
    return default


def _prompt_pin() -> int | None:
    pin = _env("ELKE27_PIN")
    pin = pin.strip() if pin else input("Panel requires authorization. Enter PIN: ").strip()

    if not pin.isdigit() or not (4 <= len(pin) <= 6):
        print("ERROR: PIN must be 4–6 digits.")
        return None
    return int(pin)


def _dict_auth_required(obj: dict) -> bool:
    net = obj.get("network")
    if not isinstance(net, dict):
        return False
    error_code = net.get("error_code")
    if isinstance(error_code, str):
        try:
            error_code = int(error_code)
        except ValueError:
            return False
    return error_code == 11008


def _wait_for_network_event(elk: Elk, kind: str, timeout_s: float):
    deadline = time.monotonic() + timeout_s
    while time.monotonic() < deadline:
        obj = elk.session.pump_once(timeout_s=0.5)
        if isinstance(obj, dict) and _dict_auth_required(obj):
            return "auth_required", None

        for evt in elk.drain_events():
            if evt.kind == kind:
                return "ok", evt
            if evt.kind == ApiError.KIND and evt.error_code == 11008 and evt.scope == "network":
                return "auth_required", evt
            if evt.kind == ApiError.KIND and evt.scope == "network":
                return "error", evt

    return "timeout", None


def _authenticate(elk: Elk, *, pin: int) -> None:
    auth_msg = {"seq": 110, "authenticate": {"pin": pin}}
    elk.session.send_json(auth_msg)
    try:
        reply = elk.session.recv_json(timeout_s=2.0)
        print(f"Authenticate reply: {reply}")
    except Exception as exc:
        print(f"Authenticate reply read failed: {exc}")


async def _run_with_auth_retry(
    elk: Elk,
    route: tuple[str, str],
    *,
    success_kind: str,
    description: str,
    timeout_s: float,
) -> bool:
    attempt = 0
    while attempt < 2:
        attempt += 1
        seq = elk.request(route)
        print(f"Sent {description} seq={seq}")

        status, evt = _wait_for_network_event(elk, success_kind, timeout_s)
        if status == "ok":
            return True
        if status == "error":
            assert isinstance(evt, ApiError)
            print(f"{description} failed with error_code={evt.error_code} message={evt.message!r}")
            return False
        if status == "timeout":
            print(f"Timed out waiting for {description} response.")
            return False

        if status == "auth_required":
            if attempt > 1:
                print("Authorization still required (PIN rejected or insufficient privileges).")
                return False
            print(f"Authorization required for {description}; authenticating…")
            pin = _prompt_pin()
            if pin is None:
                return False
            _authenticate(elk, pin=pin)
            print(f"Authenticated; retrying {description}…")

    return False


async def main_async() -> int:
    ap = argparse.ArgumentParser()
    ap.add_argument("--timeout", type=float, default=10.0, help="Overall timeout for request/response")
    ap.add_argument("--address", type=str, default=None, help="Optional discovery broadcast address override")
    ap.add_argument("--log-level", default=_env_any(["LOG_LEVEL"], "INFO"), help="DEBUG/INFO/WARNING/ERROR")
    ap.add_argument(
        "--host",
        type=str,
        default=_env_any(["ELKE27_HOST", "ELK_HOST"], None),
        help="Panel IP/host (skip discovery)",
    )
    ap.add_argument(
        "--port",
        type=int,
        default=int(_env_any(["ELKE27_PORT", "ELK_PORT"], "2101") or 2101),
        help="Panel port (default 2101)",
    )
    args = ap.parse_args()
    logging.basicConfig(level=getattr(logging, str(args.log_level).upper(), logging.INFO))

    mn = _env_any(["ELKE27_MN", "ELK_MN"], "")
    sn = _env_any(["ELKE27_SN", "ELK_SN"], "")
    access_code = _env_any(["ELKE27_ACCESS_CODE", "ELK_ACCESS_CODE"], "")
    passphrase = _env_any(["ELKE27_PASSPHRASE", "ELK_PASSPHRASE"], "")

    if not mn or not sn:
        raise SystemExit("ERROR: set ELK_MN and ELK_SN (client_identity).")
    if not access_code or not passphrase:
        raise SystemExit("ERROR: set ELK_ACCESS_CODE and ELK_PASSPHRASE (provisioning).")

    # Panel selection
    if args.host:
        panel = {"host": args.host, "port": args.port}
    else:
        discovered = await Elk.discover(timeout=int(args.timeout), address=args.address)
        if not discovered.panels:
            raise SystemExit("ERROR: no panels discovered.")
        panel = discovered.panels[0]

    elk = Elk(features=["elke27_lib.features.network_param"])

    client_identity = linking.E27Identity(
        mn=str(mn),
        sn=str(sn),
        fwver=str(_env_any(["ELKE27_FWVER", "ELK_FWVER"], "0") or "0"),
        hwver=str(_env_any(["ELKE27_HWVER", "ELK_HWVER"], "0") or "0"),
        osver=str(_env_any(["ELKE27_OSVER", "ELK_OSVER"], "0") or "0"),
    )
    creds = E27Credentials(access_code=access_code, passphrase=passphrase)

    link_keys = await elk.link(panel, client_identity, creds, timeout_s=float(args.timeout))
    if isinstance(panel, dict):
        panel_host = panel.get("panel_host") or panel.get("host") or panel.get("ip_address")
        panel_port = int(panel.get("port") or 2101)
    else:
        panel_host = panel.panel_host
        panel_port = int(panel.port)

    session_cfg = SessionConfig(host=panel_host, port=panel_port, hello_timeout_s=float(args.timeout))
    await elk.connect(link_keys, client_identity=client_identity, session_config=session_cfg)

    ok = await _run_with_auth_retry(
        elk,
        ("network", "get_ssid"),
        success_kind=NetworkSsidResultsUpdated.KIND,
        description="network.get_ssid",
        timeout_s=float(args.timeout),
    )
    if ok:
        results = elk.state.network.ssid_scan_results
        print(f"SSID results: {[r.get('ssid') for r in results if isinstance(r, dict)]}")

    ok = await _run_with_auth_retry(
        elk,
        ("network", "get_rssi"),
        success_kind=NetworkRssiUpdated.KIND,
        description="network.get_rssi",
        timeout_s=float(args.timeout),
    )
    if ok:
        print(f"RSSI: {elk.state.network.rssi}")

    await elk.close()
    return 0


def main() -> int:
    return asyncio.run(main_async())


if __name__ == "__main__":
    raise SystemExit(main())



"""Output domain request generators."""

from __future__ import annotations

ResponseKey = tuple[str, str]


def generator_output_get_table_info() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("output", "get_table_info")


def generator_output_get_configured(*, block_id: int = 1) -> tuple[dict[str, object], ResponseKey]:
    if block_id < 1:
        raise ValueError(f"block_id must be an int >= 1 (got {block_id!r})")
    return {"block_id": block_id}, ("output", "get_configured")


def generator_output_get_status(*, output_id: int) -> tuple[dict[str, object], ResponseKey]:
    if output_id < 1:
        raise ValueError(f"output_id must be an int >= 1 (got {output_id!r})")
    return {"output_id": output_id}, ("output", "get_status")


def generator_output_set_status(
    *, output_id: int, status: str
) -> tuple[dict[str, object], ResponseKey]:
    if output_id < 1:
        raise ValueError(f"output_id must be an int >= 1 (got {output_id!r})")
    if status not in {"ON", "OFF"}:
        raise ValueError(f"status must be 'ON' or 'OFF' (got {status!r})")
    return {"output_id": output_id, "status": status}, ("output", "set_status")


def generator_output_get_attribs(*, output_id: int) -> tuple[dict[str, object], ResponseKey]:
    if output_id < 1:
        raise ValueError(f"output_id must be an int >= 1 (got {output_id!r})")
    return {"output_id": output_id}, ("output", "get_attribs")


def generator_output_get_all_outputs_status(
    *, block_id: int = 1
) -> tuple[dict[str, object], ResponseKey]:
    if block_id < 1:
        raise ValueError(f"block_id must be an int >= 1 (got {block_id!r})")
    return {"block_id": block_id}, ("output", "get_all_outputs_status")


def generator_output_get_available() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("output", "get_available")



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    DispatchRoutingError,
    OutputConfiguredInventoryReady,
    OutputConfiguredUpdated,
    OutputsStatusBulkUpdated,
    OutputStatusUpdated,
    OutputTableInfoUpdated,
    TableCsmChanged,
)
from elke27_lib.handlers import output as output_handler
from elke27_lib.states import OutputState, PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_output_get_status_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = output_handler.make_output_get_status_handler(state, emit, now=lambda: 1.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"output": {}}, make_ctx()) is False

    msg = {"output": {"get_status": {"error_code": 3}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"output": {"get_status": {"output_id": 0}}}
    assert handler(msg, make_ctx()) is False

    emit.events.clear()
    msg = {"output": {"get_status": {"output_id": 1, "status": "on", "extra": 1}}}
    assert handler(msg, make_ctx()) is True
    output = state.outputs[1]
    assert output.status == "ON"
    assert output.on is True
    assert output.fields["extra"] == 1
    assert _any_event(emit, OutputStatusUpdated)
    assert state.panel.last_message_at == 1.0


def test_output_set_status_and_get_available_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    set_status = output_handler.make_output_set_status_handler(state, emit, now=lambda: 1.5)
    get_available = output_handler.make_output_get_available_handler(state, emit, now=lambda: 1.6)

    assert set_status({"nope": {}}, make_ctx()) is False
    assert set_status({"output": {}}, make_ctx()) is False
    assert set_status({"output": {"set_status": {"output_id": 0}}}, make_ctx()) is False

    msg = {"output": {"set_status": {"output_id": 1, "status": "ON"}}}
    assert set_status(msg, make_ctx()) is True
    assert state.outputs[1].on is True
    assert _any_event(emit, OutputStatusUpdated)

    assert get_available({"nope": {}}, make_ctx()) is False
    assert get_available({"output": {}}, make_ctx()) is False
    msg = {"output": {"get_available": {"outputs": [1, 2]}}}
    assert get_available(msg, make_ctx()) is True
    assert state.inventory.configured_outputs == {1, 2}
    assert state.panel.last_message_at == 1.6


def test_output_get_configured_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = output_handler.make_output_get_configured_handler(state, emit, now=lambda: 2.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"output": {}}, make_ctx()) is False

    msg = {"output": {"get_configured": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"output": {"get_configured": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    state.table_info_by_domain["output"] = {"table_elements": 2}
    msg = {"output": {"get_configured": {"outputs": [1, 2, 3]}}}
    assert handler(msg, make_ctx()) is True
    assert state.inventory.configured_outputs == {1, 2}
    assert state.inventory.configured_outputs_complete is True
    assert _any_event(emit, OutputConfiguredUpdated)
    assert _any_event(emit, OutputConfiguredInventoryReady)
    assert state.panel.last_message_at == 2.0


def test_output_get_all_outputs_status_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = output_handler.make_output_get_all_outputs_status_handler(
        state, emit, now=lambda: 3.0
    )

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"output": {}}, make_ctx()) is False

    msg = {"output": {"get_all_outputs_status": {"error_code": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"output": {"get_all_outputs_status": {"foo": "bar"}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, DispatchRoutingError)

    emit.events.clear()
    msg = {"output": {"get_all_outputs_status": {"status": "1 0 1x"}}}
    assert handler(msg, make_ctx()) is True
    assert state.outputs[1].on is True
    assert state.outputs[2].on is False
    assert state.outputs[3].on is True
    assert _any_event(emit, OutputsStatusBulkUpdated)
    assert state.panel.last_message_at == 3.0


def test_output_configured_merge_and_attribs() -> None:
    state = PanelState()
    merge = output_handler.make_output_configured_merge(state)
    merged = merge(
        [
            PagedBlock(block_id=1, payload={"outputs": [1, 2]}),
            PagedBlock(block_id=2, payload={"output_ids": [2, 3]}),
        ],
        2,
    )
    assert merged == {"outputs": [1, 2, 3], "block_count": 2}

    emit = _EmitSpy()
    attribs = output_handler.make_output_get_attribs_handler(state, emit, now=lambda: 4.0)
    assert attribs({"nope": {}}, make_ctx()) is False
    assert attribs({"output": {}}, make_ctx()) is False

    msg = {"output": {"get_attribs": {"error_code": 11008, "output_id": 1}}}
    assert attribs(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"output": {"get_attribs": {"error_code": 9, "output_id": 1}}}
    assert attribs(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"output": {"get_attribs": {"output_id": 0}}}
    assert attribs(msg, make_ctx()) is False

    msg = {"output": {"get_attribs": {"output_id": 1, "name": "  Test "}}}
    assert attribs(msg, make_ctx()) is True
    assert state.outputs[1].name == "Test"
    assert state.panel.last_message_at == 4.0


def test_output_get_table_info_handler_and_helpers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = output_handler.make_output_get_table_info_handler(state, emit, now=lambda: 5.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"output": {}}, make_ctx()) is False

    msg = {"output": {"get_table_info": {"error_code": 3}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    state.table_info_known.update({"area", "zone", "tstat"})
    msg = {"output": {"table_info": {"table_elements": 2, "table_csm": "4"}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, OutputTableInfoUpdated)
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)
    assert _any_event(emit, BootstrapCountsReady)
    assert state.table_info_by_domain["output"]["table_elements"] == 2

    assert output_handler._extract_table_csm({"table_csm": True}, domain="output") is None
    assert output_handler._extract_table_csm({"table_csm": 2}, domain="output") == 2
    assert output_handler._extract_table_csm({"table_csm": 2.0}, domain="output") == 2
    assert output_handler._extract_table_csm({"table_csm": "3"}, domain="output") == 3
    assert output_handler._extract_table_csm({"table_csm": "bad"}, domain="output") is None
    assert output_handler._extract_table_csm({"other": 1}, domain="output") is None


def test_output_helper_functions() -> None:
    output = OutputState(output_id=1)
    changed: set[str] = set()
    output_handler._apply_output_status_fields(output, {"status": " off ", "extra": 1}, changed)
    assert output.status == "OFF"
    assert output.on is False
    assert output.fields["extra"] == 1
    assert output_handler._extract_configured_output_ids({"outputs": [1, 2, 2, 0, "x"]}) == [
        1,
        2,
    ]
    assert output_handler._extract_configured_output_ids({"output_ids": [3]}) == [3]
    assert output_handler._extract_configured_output_ids({"configured_outputs": [4]}) == [4]
    assert output_handler._extract_configured_output_ids({"configured_output_ids": [5]}) == [5]
    assert output_handler._extract_configured_output_ids({"none": []}) == []

    assert output_handler._apply_output_status_char(output, "x") is False
    assert output_handler._apply_output_status_char(output, "1") is True
    assert output.status == "ON"
    assert output.on is True

    assert output_handler._normalize_name(None) is None
    assert output_handler._normalize_name("   ") is None
    assert output_handler._normalize_name(" Name ") == "Name"

    assert output_handler._extract_int({"table_elements": 2}, "table_elements") == 2
    assert output_handler._extract_int({"table_elements": "bad"}, "table_elements") is None



"""
Live E27 test: paged read-only command (zone_get_configured).

Run:
  source ~/elk-e27-env-vars.sh
  pytest -q -m live_e27 test/test_live_e27_paged_readonly.py -s
"""

from __future__ import annotations

from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.const import E27ErrorCode
from test.helpers.error_codes import describe_error, extract_error_code
from test.helpers.payload_validation import assert_payload_shape

_LIVE_TIMEOUT_S = 60.0


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_paged_zone_get_configured(live_e27_client: Elke27Client) -> None:
    result = await live_e27_client.async_execute("zone_get_configured", timeout_s=_LIVE_TIMEOUT_S)
    if not result.ok:
        if extract_error_code(result.error) == E27ErrorCode.ELKERR_NOT_READY:
            pytest.skip("Panel reported not-ready for zone_get_configured.")
        pytest.fail(f"zone_get_configured failed: {describe_error(result.error)}")
    assert_payload_shape("zone_get_configured", result.data)

    assert isinstance(result.data, dict)
    zones = result.data.get("zones")
    block_count = result.data.get("block_count")
    assert isinstance(zones, list)
    zones_list = cast(list[object], zones)
    assert len(zones_list) > 0
    assert isinstance(block_count, int)
    if block_count <= 1:
        pytest.skip("zone_get_configured returned a single block; paging not exercised.")



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Iterable
from enum import Enum

import pytest
from _pytest.monkeypatch import MonkeyPatch

from elke27_lib import (
    AreaState,
    ArmMode,
    ClientConfig,
    CsmSnapshot,
    DiscoveredPanel,
    Elke27Client,
    Elke27Event,
    EventType,
    LinkKeys,
    OutputDefinition,
    OutputState,
    PanelInfo,
    PanelSnapshot,
    TableInfo,
    ZoneDefinition,
    ZoneState,
    redact_for_diagnostics,
)
from elke27_lib.discovery import E27System
from elke27_lib.errors import E27ProvisioningTimeout, Elke27AuthError
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    AreaConfiguredInventoryReady,
    OutputConfiguredInventoryReady,
    ZoneConfiguredInventoryReady,
)
from elke27_lib.kernel import DiscoverResult
from elke27_lib.linking import E27LinkKeys
from elke27_lib.states import InventoryState


def test_public_api_types_are_dataclasses_or_enums() -> None:
    assert dataclasses.is_dataclass(ClientConfig)
    assert dataclasses.is_dataclass(DiscoveredPanel)
    assert dataclasses.is_dataclass(LinkKeys)
    assert dataclasses.is_dataclass(Elke27Event)
    assert dataclasses.is_dataclass(PanelInfo)
    assert dataclasses.is_dataclass(TableInfo)
    assert dataclasses.is_dataclass(CsmSnapshot)
    assert dataclasses.is_dataclass(AreaState)
    assert dataclasses.is_dataclass(ZoneState)
    assert dataclasses.is_dataclass(OutputState)
    assert dataclasses.is_dataclass(PanelSnapshot)
    assert dataclasses.is_dataclass(ZoneDefinition)
    assert dataclasses.is_dataclass(OutputDefinition)
    assert issubclass(EventType, Enum)
    assert issubclass(ArmMode, Enum)


def test_public_api_client_instantiation_and_snapshot() -> None:
    client = Elke27Client()
    snapshot = client.snapshot
    assert isinstance(snapshot, PanelSnapshot)
    assert snapshot.version == 0


def test_redaction_returns_json_serializable() -> None:
    data = {"access_code": "1234", "nested": {"token": "abc"}, "count": 1}
    redacted = redact_for_diagnostics(data)
    assert redacted["access_code"] == "***"
    assert redacted["nested"]["token"] == "***"
    assert redacted["count"] == 1


@pytest.mark.asyncio
async def test_async_discover_maps_panels(monkeypatch: MonkeyPatch) -> None:
    async def _fake_discover(*_args: object, **_kwargs: object) -> DiscoverResult:
        return DiscoverResult(
            panels=[
                E27System(
                    panel_mac="aa:bb:cc",
                    panel_host="1.2.3.4",
                    panel_name="Panel",
                    panel_serial="SN",
                    port=2101,
                    tls_port=2601,
                )
            ]
        )

    monkeypatch.setattr("elke27_lib.kernel.E27Kernel.discover", _fake_discover)
    client = Elke27Client()
    panels = await client.async_discover()
    assert len(panels) == 1
    assert panels[0].host == "1.2.3.4"
    assert panels[0].port == 2101
    assert panels[0].tls_port == 2601


@pytest.mark.asyncio
async def test_async_link_roundtrip_and_auth_error(monkeypatch: MonkeyPatch) -> None:
    client = Elke27Client()

    async def _fake_link(*_args: object, **_kwargs: object) -> E27LinkKeys:
        return E27LinkKeys(tempkey_hex="t", linkkey_hex="k", linkhmac_hex="h")

    kernel = client._kernel
    monkeypatch.setattr(kernel, "link", _fake_link)
    keys = await client.async_link(
        "1.2.3.4",
        2101,
        access_code="1234",
        passphrase="pass",
        client_identity={"mn": "222", "sn": "00000001"},
    )
    assert keys.to_json() == {"tempkey_hex": "t", "linkkey_hex": "k", "linkhmac_hex": "h"}
    assert LinkKeys.from_json(keys.to_json()) == keys

    async def _fake_link_error(*_args: object, **_kwargs: object) -> E27LinkKeys:
        raise E27ProvisioningTimeout("auth failed")

    monkeypatch.setattr(kernel, "link", _fake_link_error)
    with pytest.raises(Elke27AuthError):
        await client.async_link(
            "1.2.3.4",
            2101,
            access_code="1234",
            passphrase="pass",
            client_identity={"mn": "222", "sn": "00000001"},
        )


@pytest.mark.asyncio
async def test_async_connect_disconnect_and_wait_ready(monkeypatch: MonkeyPatch) -> None:
    del monkeypatch

    class _FakeKernel:
        _ready: bool
        state: object

        def __init__(self) -> None:
            self._ready = False
            self.state = type(
                "_State",
                (),
                {
                    "panel": type(
                        "_Panel", (), {"model": None, "firmware": None, "serial": None}
                    )(),
                    "table_info_by_domain": {
                        "area": {"table_elements": None},
                        "zone": {"table_elements": None},
                        "output": {"table_elements": None},
                        "tstat": {"table_elements": None},
                    },
                    "areas": {},
                    "zones": {},
                    "outputs": {},
                    "inventory": InventoryState(),
                },
            )()

        @property
        def ready(self) -> bool:
            return self._ready

        async def connect(self, *_args: object, **_kwargs: object) -> None:
            self._ready = True

        async def close(self) -> None:
            self._ready = False

        def subscribe(
            self, _callback: Callable[[object], None], _kinds: Iterable[str] | None = None
        ) -> int:
            return 1

        def unsubscribe(self, _token: int) -> bool:
            return True

        def drain_events(self) -> list[object]:
            return []

    client = Elke27Client(kernel=_FakeKernel())
    await client.async_connect(
        "1.2.3.4", 2101, LinkKeys(tempkey_hex="t", linkkey_hex="k", linkhmac_hex="h")
    )
    assert client.is_ready is False
    handle_event = client._handle_kernel_event
    handle_event(
        AreaConfiguredInventoryReady(
            kind=AreaConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    handle_event(
        ZoneConfiguredInventoryReady(
            kind=ZoneConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    handle_event(
        OutputConfiguredInventoryReady(
            kind=OutputConfiguredInventoryReady.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
        )
    )
    assert await client.wait_ready(timeout_s=0.01) is True
    await client.async_disconnect()
    assert client.is_ready is False



#!/usr/bin/env python3
"""
Minimal client contract example for Elke27Client.

Flow:
- discover or use --host/--port
- link -> connect
- wait ready
- subscribe and receive an event
- read one state snapshot
"""

from __future__ import annotations

import argparse
import asyncio
import logging
import time
from dataclasses import dataclass

from elke27_lib import Elke27Client

LOG = logging.getLogger(__name__)


@dataclass(frozen=True, slots=True)
class E27Identity:
    mn: str
    sn: str
    fwver: str = "0"
    hwver: str = "0"
    osver: str = "0"


@dataclass(frozen=True, slots=True)
class E27Credentials:
    access_code: str
    passphrase: str


def _env(name: str, default: str | None = None) -> str | None:
    import os

    return os.environ.get(name, default)


def _env_any(names: list[str], default: str | None = None) -> str | None:
    for name in names:
        value = _env(name)
        if value:
            return value
    return default


def _truthy(value: str | None) -> bool:
    if value is None:
        return False
    return str(value).strip().lower() in {"1", "true", "yes", "on"}


async def main_async() -> int:
    ap = argparse.ArgumentParser()
    ap.add_argument("--timeout", type=float, default=10.0, help="Overall timeout")
    ap.add_argument("--pump-s", type=float, default=0.25, help="Per-pump timeout")
    ap.add_argument("--address", type=str, default=None, help="Discovery broadcast override")
    ap.add_argument("--host", type=str, default=_env_any(["ELKE27_HOST", "ELK_HOST"], None))
    ap.add_argument("--port", type=int, default=int(_env_any(["ELKE27_PORT", "ELK_PORT"], "2101") or 2101))
    ap.add_argument("--log-level", default=_env_any(["LOG_LEVEL"], "INFO"))
    args = ap.parse_args()

    logging.basicConfig(level=getattr(logging, str(args.log_level).upper(), logging.INFO))

    mn = _env_any(["ELKE27_MN", "ELK_MN"], "")
    sn = _env_any(["ELKE27_SN", "ELK_SN"], "")
    access_code = _env_any(["ELKE27_ACCESS_CODE", "ELK_ACCESS_CODE"], "")
    passphrase = _env_any(["ELKE27_PASSPHRASE", "ELK_PASSPHRASE"], "")

    if not mn or not sn:
        raise SystemExit("ERROR: set ELK_MN and ELK_SN (client_identity).")
    if not access_code or not passphrase:
        raise SystemExit("ERROR: set ELK_ACCESS_CODE and ELK_PASSPHRASE (provisioning).")

    client = Elke27Client()

    if args.host:
        panel = {"host": args.host, "port": args.port}
    else:
        discover = await client.discover(timeout=int(args.timeout), address=args.address)
        if not discover.ok or not discover.data or not discover.data.panels:
            raise SystemExit("ERROR: no panels discovered.")
        panel = discover.data.panels[0]

    client_identity = E27Identity(mn=mn, sn=sn, fwver="0", hwver="0", osver="0")
    creds = E27Credentials(access_code=access_code, passphrase=passphrase)

    link = await client.link(panel, client_identity, creds, timeout_s=float(args.timeout))
    if not link.ok:
        raise SystemExit(f"ERROR: link failed: {link.error}")

    connect = await client.connect(link.data, panel=panel, client_identity=client_identity)
    if not connect.ok:
        raise SystemExit(f"ERROR: connect failed: {connect.error}")

    ready = await asyncio.to_thread(client.wait_ready, timeout_s=float(args.timeout))
    if not ready:
        raise SystemExit("ERROR: timed out waiting for readiness.")

    got_event = False

    def _on_event(event) -> None:
        nonlocal got_event
        got_event = True
        LOG.info("Event received: %s", event.kind)

    client.subscribe(_on_event)
    request = await asyncio.to_thread(client.request, ("control", "get_version_info"))
    if not request.ok:
        raise SystemExit(f"ERROR: request failed: {request.error}")

    deadline = time.monotonic() + float(args.timeout)
    while time.monotonic() < deadline and not got_event:
        pump = await asyncio.to_thread(client.pump_once, timeout_s=float(args.pump_s))
        if not pump.ok:
            raise SystemExit(f"ERROR: pump failed: {pump.error}")
        for event in client.drain_events():
            LOG.debug("Drained event: %s", event.kind)
        await asyncio.sleep(0)

    client.unsubscribe(_on_event)

    print("Snapshot:")
    panel_info = client.panel_info
    print(f"  session_id = {panel_info.session_id}")
    print(f"  model      = {panel_info.model}")
    print(f"  firmware   = {panel_info.firmware}")
    print(f"  serial     = {panel_info.serial}")

    close = await client.disconnect()
    if not close.ok:
        raise SystemExit(f"ERROR: close failed: {close.error}")

    return 0


def main() -> int:
    return asyncio.run(main_async())


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext, MessageKind
from elke27_lib.handlers.zone import make_zone_get_all_zones_status_handler
from elke27_lib.states import PanelState


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, ctx: DispatchContext) -> None:
        self.events.append(evt)


def _ctx(route: tuple[str, str]) -> DispatchContext:
    return DispatchContext(
        kind=MessageKind.DIRECTED,
        seq=None,
        session_id=None,
        route=route,
        classification="RESPONSE",
    )


def test_bulk_status_respects_configured_zones() -> None:
    state = PanelState()
    state.inventory.configured_zones = {1, 2, 3}
    state.get_or_create_zone(1)
    emit = _EmitSpy()
    handler = make_zone_get_all_zones_status_handler(state, emit, now=lambda: 123.0)

    msg = {"zone": {"get_all_zones_status": {"status": "33333", "error_code": 0}}}
    assert handler(msg, _ctx(("zone", "get_all_zones_status"))) is True
    assert set(state.zones.keys()) == {1}



from __future__ import annotations

import logging

from elke27_lib.dispatcher import DispatchContext, PagedBlock
from elke27_lib.events import (
    ApiError,
    AreaAttribsUpdated,
    AreaConfiguredInventoryReady,
    AreaConfiguredUpdated,
    AreaStatusUpdated,
    AreaTableInfoUpdated,
    AreaTroublesUpdated,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    DispatchRoutingError,
    TableCsmChanged,
    UnknownMessage,
)
from elke27_lib.handlers import area as area_handler
from elke27_lib.states import AreaState, PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


class _EmitRaiser:
    def __call__(self, _evt: object, _ctx: DispatchContext) -> None:
        raise RuntimeError("emit failed")


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_reconcile_area_state_and_helpers() -> None:
    state = PanelState()
    outcome = area_handler._reconcile_area_state(state, {"area_id": 0}, now=1.0, _source="test")
    assert outcome.area_id == -1
    assert outcome.warnings

    outcome = area_handler._reconcile_area_state(
        state,
        {"area_id": 1, "arm_state": "ARMED", "ready": "bad", "num_bypassed_zones": 2},
        now=2.0,
        _source="test",
    )
    assert outcome.area_id == 1
    assert "arm_state" in outcome.changed_fields
    assert "num_bypassed_zones" in outcome.changed_fields
    assert outcome.warnings

    assert area_handler._coerce_intish(2) == 2
    assert area_handler._coerce_intish("3") == 3
    assert area_handler._coerce_intish("bad") is None
    assert area_handler._has_configured_area_ids({"areas": []}) is True
    assert area_handler._has_configured_area_ids({"other": 1}) is False

    warnings: list[str] = []
    assert area_handler._parse_area_id_container(None, warnings) == []
    assert area_handler._parse_area_id_container([1, "2", 0], warnings) == [1, 2]
    assert area_handler._parse_area_id_container({"1": True, "2": False}, warnings) == [1]
    assert area_handler._parse_area_id_container({"x": {"area_id": 3}}, warnings) == [3]
    assert area_handler._parse_area_id_container({"x": "bad"}, warnings) == []
    assert area_handler._parse_area_id_container(0x3, warnings) == [1, 2]
    assert area_handler._parse_area_id_container("0x2", warnings) == [2]
    assert area_handler._parse_area_id_container("zz", warnings) == []
    assert area_handler._parse_area_id_container(object(), warnings) == []

    assert area_handler._ids_from_bitmask(0b101) == [1, 3]
    assert area_handler._coerce_area_id(" 4 ") == 4
    assert area_handler._coerce_area_id("   ") is None
    assert area_handler._coerce_area_id("bad") is None
    assert area_handler._coerce_area_id(-1) is None
    assert area_handler._dedupe_sorted([2, 1, 1]) == [1, 2]

    assert area_handler._extract_troubles_list({"troubles": ["a", None]}) == ["a"]
    assert area_handler._extract_troubles_list({"trouble": "x"}) == ["x"]
    assert area_handler._extract_troubles_list({"list": " "}) == []
    assert area_handler._extract_troubles_list({"none": 1}) == []

    assert area_handler._normalize_name(None) is None
    assert area_handler._normalize_name("   ") is None
    assert area_handler._normalize_name(" Name ") == "Name"

    area = AreaState(area_id=1)
    changed: set[str] = set()
    area_handler._apply_area_attribs(area, {"name": " Main "}, changed)
    assert area.name == "Main"
    assert "name" in changed

    assert area_handler._extract_int({"table_elements": 2}, "table_elements") == 2
    assert area_handler._extract_int({"table_elements": "bad"}, "table_elements") is None
    assert area_handler._type_name((int, str)) == "int | str"


def test_area_get_status_handler_paths() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = area_handler.make_area_get_status_handler(state, emit, now=lambda: 3.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"area": {}}, make_ctx()) is False

    msg = {"area": {"get_status": {"error_code": 4, "area_id": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    state.debug_last_raw_by_route_enabled = True
    msg = {"area": {"get_status": {"area_id": 0}}}
    assert handler(msg, make_ctx()) is False
    assert _any_event(emit, DispatchRoutingError)
    assert "area.get_status" in state.debug_last_raw_by_route

    emit.events.clear()
    msg = {"area": {"get_status": {"area_id": 1, "ready": "bad"}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, DispatchRoutingError)

    emit.events.clear()
    msg = {"area": {"get_status": {"area_id": 1, "ready": True}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AreaStatusUpdated)

    # emit failure path
    handler_raise = area_handler.make_area_get_status_handler(state, _EmitRaiser(), now=lambda: 4.0)
    assert handler_raise({"area": {"get_status": {"area_id": 1}}}, make_ctx()) is True


def test_area_get_attribs_handler_paths() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = area_handler.make_area_get_attribs_handler(state, emit, now=lambda: 5.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"area": {}}, make_ctx()) is False

    msg = {"area": {"get_attribs": {"error_code": 11008, "area_id": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    inv = state.inventory
    inv.invalid_id_streak_threshold = 2
    inv.configured_areas = {1, 2, 3}
    inv.area_attribs_requested = {1, 2, 3}
    msg = {"area": {"get_attribs": {"error_code": 11006, "area_id": 1}}}
    assert handler(msg, make_ctx()) is True
    msg = {"area": {"get_attribs": {"error_code": 11006, "area_id": 2}}}
    assert handler(msg, make_ctx()) is True
    assert inv.area_discovery_max_id == 0
    assert inv.configured_areas == set()
    assert inv.area_attribs_requested == set()

    emit.events.clear()
    msg = {"area": {"get_attribs": {"error_code": 9, "area_id": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"area": {"get_attribs": {"area_id": 0}}}
    assert handler(msg, make_ctx()) is False

    emit.events.clear()
    msg = {"area": {"get_attribs": {"area_id": 1, "name": "  Home "}}}
    assert handler(msg, make_ctx()) is True
    assert state.areas[1].name == "Home"
    assert _any_event(emit, AreaAttribsUpdated)
    assert state.panel.last_message_at == 5.0


def test_area_get_configured_handler_and_merge() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = area_handler.make_area_get_configured_handler(state, emit, now=lambda: 6.0)
    logging.getLogger("elke27_lib.handlers.area").setLevel(logging.DEBUG)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"area": {}}, make_ctx()) is False

    msg = {"area": {"get_configured": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"area": {"get_configured": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"area": {"get_configured": {"block_id": 0, "block_count": 1}}}
    assert handler(msg, make_ctx()) is True

    msg = {"area": {"get_configured": {"block_id": 1, "block_count": 0}}}
    assert handler(msg, make_ctx()) is True

    emit.events.clear()
    msg = {"area": {"get_configured": {"areas": [1, 2], "block_id": 1, "block_count": 1}}}
    assert handler(msg, make_ctx()) is True
    assert state.inventory.configured_areas == {1, 2}
    assert _any_event(emit, AreaConfiguredUpdated)
    assert _any_event(emit, AreaConfiguredInventoryReady)

    emit.events.clear()
    msg = {"area": {"get_configured": {"block_id": 1, "block_count": 1, "bitmask": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, DispatchRoutingError)

    merge = area_handler.make_area_configured_merge(state)
    merged = merge(
        [
            PagedBlock(block_id=1, payload={"areas": [1], "block_size": 1}),
            PagedBlock(block_id=2, payload={"areas": [1], "block_size": 1}),
        ],
        2,
    )
    assert merged["areas"] == [1, 2]


def test_area_set_status_and_troubles() -> None:
    state = PanelState()
    emit = _EmitSpy()
    set_status = area_handler.make_area_set_status_handler(state, emit, now=lambda: 7.0)

    assert set_status({"nope": {}}, make_ctx()) is False
    assert set_status({"area": {}}, make_ctx()) is False

    msg = {"area": {"set_status": {"area_id": 0}}}
    assert set_status(msg, make_ctx()) is False
    assert _any_event(emit, DispatchRoutingError)

    emit.events.clear()
    msg = {"area": {"set_status": {"area_id": 1, "arm_state": "ARMED", "error_code": 2}}}
    assert set_status(msg, make_ctx()) is True
    assert _any_event(emit, AreaStatusUpdated)
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"area": {"set_status": {"area_id": 1, "ready": "bad"}}}
    assert set_status(msg, make_ctx()) is True
    assert _any_event(emit, DispatchRoutingError)

    troubles = area_handler.make_area_get_troubles_handler(state, emit, now=lambda: 8.0)
    assert troubles({"nope": {}}, make_ctx()) is False
    assert troubles({"area": {}}, make_ctx()) is False

    msg = {"area": {"get_trouble": {"error_code": 3}}}
    assert troubles(msg, make_ctx(classification="BROADCAST")) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"area": {"get_troubles": {"area_id": 0}}}
    assert troubles(msg, make_ctx()) is False

    emit.events.clear()
    msg = {"area": {"get_troubles": {"area_id": 1, "troubles": ["x"]}}}
    assert troubles(msg, make_ctx()) is True
    assert _any_event(emit, AreaTroublesUpdated)
    assert state.areas[1].troubles == ["x"]


def test_area_get_table_info_and_root_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = area_handler.make_area_get_table_info_handler(state, emit, now=lambda: 9.0)
    logging.getLogger("elke27_lib.handlers.area").setLevel(logging.DEBUG)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"area": {}}, make_ctx()) is False

    msg = {"area": {"get_table_info": {"error_code": 1}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    state.table_info_known.update({"zone", "output", "tstat"})
    state.inventory.configured_areas = {1, 2, 3}
    state.inventory.area_attribs_requested = {1, 2, 3}
    msg = {"area": {"table_info": {"table_elements": 2, "table_csm": "5"}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AreaTableInfoUpdated)
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, CsmSnapshotUpdated)
    assert _any_event(emit, AreaConfiguredUpdated)
    assert _any_event(emit, BootstrapCountsReady)
    assert state.inventory.configured_areas == {1, 2}

    state.debug_last_raw_by_route_enabled = True
    root_handler = area_handler.make_area___root___handler(state, emit, lambda: 0.0)
    msg = {"area": {"foo": "bar", "baz": "qux"}}
    assert root_handler(msg, make_ctx()) is True
    assert _any_event(emit, UnknownMessage)
    assert "area.__root__" in state.debug_last_raw_by_route

    assert root_handler({"nope": {}}, make_ctx()) is False


def test_area_configured_helpers() -> None:
    state = PanelState()
    import logging

    logging.getLogger("elke27_lib.handlers.area").setLevel(logging.DEBUG)
    warnings: list[str] = []
    payload = {"configured": [1, 2], "block_id": 1, "block_count": 1}
    outcome = area_handler._reconcile_configured_areas(state, payload, now=10.0)
    assert outcome.configured_ids == (1, 2)

    state.table_info_by_domain["area"] = {"table_elements": 1}
    outcome = area_handler._reconcile_configured_areas(state, payload, now=10.0)
    assert outcome.configured_ids == (1,)

    warnings.clear()
    ids = area_handler._extract_configured_area_ids({"bitmask": 0x3}, warnings)
    assert ids == [1, 2]
    assert "no configured area ids found" not in warnings

    warnings.clear()
    ids = area_handler._extract_configured_area_ids(
        {"block_id": 1, "block_count": 2, "bitmask": 1}, warnings
    )
    assert ids == []
    assert "bitmask ignored" in warnings[0]

    block_size = area_handler._configured_block_size({"block_size": 2}, state, 2, domain="area")
    assert block_size == 2
    state.table_info_by_domain["area"] = {"table_elements": 5}
    assert area_handler._configured_block_size({}, state, 2, domain="area") == 3
    state.table_info_by_domain["area"] = {"table_elements": 0}
    assert area_handler._configured_block_size({}, state, 2, domain="area") is None

    assert area_handler._apply_configured_block_offset([1], block_id=2, block_size=2) == [3]
    assert area_handler._apply_configured_block_offset([], block_id=2, block_size=2) == []
    assert area_handler._apply_configured_block_offset([3], block_id=2, block_size=2) == [3]
    assert area_handler._apply_configured_block_offset([1], block_id=2, block_size=None) == [1]

    assert area_handler._extract_table_csm({"table_csm": True}, domain="area") is None
    assert area_handler._extract_table_csm({"table_csm": 2}, domain="area") == 2
    assert area_handler._extract_table_csm({"table_csm": 2.0}, domain="area") == 2
    assert area_handler._extract_table_csm({"table_csm": "3"}, domain="area") == 3
    assert area_handler._extract_table_csm({"table_csm": "bad"}, domain="area") is None
    assert area_handler._extract_table_csm({"other": 1}, domain="area") is None
